import asyncio
import datetime
import time

from AnkiIn.config import dict as conf
from AnkiIn.helper import ankiConnectHelper as ankiHelper
from ankisiyuan import config, config_parser
from ankisiyuan.parsers import siyuan as siyuan_parser


updated_notes = 0
added_notes = 0


async def execute():
    if not ankiHelper.check_online():
        print(
            "Anki is not online! Please open Anki and \
make sure anki-connect is installed."
        )
        return
    start_time = time.perf_counter()
    config_parser.parse()
    print(f'Api Token:{conf["siyuan"]["api_token"]}')
    print(f"Last Sync:{config.last_sync_time}")
    this_sync_time = datetime.datetime.now().strftime(r"%Y%m%d%H%M%S")
    print(f"Now Sync: {this_sync_time}")
    note_list = await siyuan_parser.sync(config.last_sync_time)
    for x in note_list:
        if x is None:
            continue
        handle(x)
    print(f"Added {added_notes} notes, updated {updated_notes} notes.")
    end_time = time.perf_counter()
    print(f"Time consumed: {end_time - start_time}s")
    with open("last_sync_time", "w", encoding="UTF-8") as f:
        f.write(this_sync_time)
        f.close()


def handle(x):
    global updated_notes, added_notes
    old = ankiHelper.find_notes(f"SiyuanID:{x['SiyuanID']}*")
    # print(x.outputfields['SiyuanID'])
    # print(x.SiyuanID)
    if len(old) >= 2:
        print(
            f'found SiyuanID[{x["SiyuanID"]}] with \
            duplicates in Anki!\n{old}'
        )
    elif len(old) == 1:
        ankiHelper.update_note_fields(old[0], x)
        print(f'updated: SiyuanID[{x["SiyuanID"]}] AnkiID[{old[0]}]')
        updated_notes = updated_notes + 1
    else:
        ankiHelper.add_note(x)
        added_notes = added_notes + 1


def execute_from_commandline():
    asyncio.run(execute())



from AnkiIn.config import config_updater, dict as conf
from AnkiIn.log import notetype_logger as log
from AnkiIn.model import Model
from AnkiIn.note import Note


notetype_name = "SWatch"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = 1
enable = False


def update_swatch_config() -> None:
    global settings, priority, enable

    priority = settings.get("priority", 1)
    enable = settings.get("enable", False)


config_updater.append((update_swatch_config, 10))


def check(lines: list, extra_params: None | dict = None) -> bool:
    return enable and len(lines) >= 1


def get(text: str, deck: str, tags: list, extra_params: None | dict = None) -> Note:
    return SWatchNote(
        front=text, siyuan_id=extra_params["SiyuanID"], deck=deck, tags=tags
    )


BACK = r"""{{FrontSide}}
<hr id=answer>
<a href="siyuan://blocks/{{SiyuanID}}">Open in SiYuan</a>
"""

CSS = r""".card {
  font-family: arial;
  font-size: 20px;
  text-align: center;
  color: black;
  background-color: white;
}
ul {
display: inline-block;
text-align: left;
}
ol {
display: inline-block;
text-align: left;
}
"""

MODELNAME = "SAnkiLink-Watch"
MODELID = 1195191921

_model = Model(
    modelId=MODELID,
    modelName=MODELNAME,
    fields=["Front", "SiyuanID"],
    templates=[{"Name": "Card 1", "Front": "{{Front}}", "Back": BACK}],
    css=CSS,
)


class SWatchNote(Note):
    def __init__(self, front, siyuan_id, deck, tags):
        global _model
        super().__init__(
            model=_model,
            fields={"Front": front, "SiyuanID": siyuan_id},
            deck=deck,
            tags=tags,
        )



from __future__ import annotations

from AnkiIn import config
from AnkiIn.config import config_updater, dict as conf, update_config
from AnkiIn.helper.formatHelper import remove_suffix
from AnkiIn.log import parser_logger as logger
from AnkiIn.notetype_loader import discovered_notetypes
from AnkiIn.parser import markdown
from ankisiyuan.notetypes import SMQA, SQA, SCloze, SListCloze, STableCloze, SWatch
from siyuanhelper import api


# class SyntaxNode:
#     def __init__(self, id: str, parent: None | str = None, sons: None | list = None):
#         self.id = id
#         self.parent = parent
#         self.sons = sons
#         if self.sons is None:
#             self.sons = []


# link = {}
# is_added = {}
# roots = []
# note_list = []


# async def build_tree(now: str, siyuan: api.Siyuan):
#     # print("visit:{}".format(now))
#     # print("build tree")
#     # print(get_col_by_id(now, "markdown"))
#     now_node = SyntaxNode(now)
#     block = await siyuan.get_block_by_id(now)
#     link[now] = now_node
#     try:
#         if await block.attrs.get(tag_attr_name) != "":
#             roots.append(now_node)
#             return now_node
#     except Exception:
#         logger.exception(f"Exception occurred! Invalid Siyuan ID {now}")
#         logger.exception(Exception)
#     fa_id = await siyuan.get_parent_id_by_id(now)
#     if fa_id == "":
#         return now_node
#     # print("son: {} fa:{}".format(now, fa_id))
#     fa = link.get(fa_id)
#     if fa is None:
#         fa = await build_tree(fa_id, siyuan)
#     now_node.parent = fa
#     # print("fa {} added son {}".format(fa.id, now_node.id))
#     fa.sons.append(now_node)
#     # print("fa {} sons:".format(fa.id))
#     # print([x.id for x in fa.sons])
#     return now_node


# async def sync(last_time: str):
#     # session = aiohttp.ClientSession()
#     # set_session(session)
#     global siyuan
#     siyuan = api.Siyuan(token=conf["siyuan"].get("api_token", ""))
#     link.clear()
#     is_added.clear()
#     roots.clear()
#     note_list.clear()
#     all_origin_blocks = await siyuan.sql_query(
#         f"SELECT id FROM blocks where updated>'{last_time}' and type='p'"
#     )
#     all_blocks = [x["id"] for x in all_origin_blocks]
#     # print(all_blocks)
#     tasks = []
#     for x in all_blocks:
#         tasks.append(asyncio.create_task(build_tree(x, siyuan)))
#     await asyncio.tasks.gather(*tasks)
#     # print([x.id for x in roots])
#     # print([get_col_by_id(x.id,"markdown") for x in roots])
#     for x in roots:
#         await dfs(x, siyuan)
#     await siyuan.close()
#     return note_list


# async def dfs(now: SyntaxNode, siyuan: api.Siyuan):
#     # print("dfs: " + now.id)
#     # print([x.id for x in now.sons])
#     current_config = None
#     config_backup = None
#     try:
#         attrs = await siyuan.get_attrs_by_id(now.id)
#         if "custom-ankilink" not in attrs:
#             logger.debug(f"Siyuan Block {now.id} has no ankilink property.")
#         else:
#             current_config = attrs.get(tag_attr_name)
#             config_backup = config.parse_config(current_config)
#     except Exception as e:
#         logger.warning(
#             f"An error occurred while parsing config.\nSiyuanID:{now.id}\nProperty:\n{current_config}\nExpcetion:{e.with_traceback()}"
#         )
#     if len(now.sons) == 0:
#         await handle(now, siyuan)
#     else:
#         for x in now.sons:
#             await dfs(x, siyuan)
#     if config_backup is not None:
#         config.execute_config(config_backup)


# async def handle(now: SyntaxNode, siyuan: api.Siyuan):
#     fa = await siyuan.get_parent_id_by_id(now.id)
#     if fa == "":
#         await add_note(now.id, siyuan)
#         return
#     now_block = await siyuan.get_block_by_id(now.id)
#     fa_block = await siyuan.get_block_by_id(fa)
#     await now_block.ensure()
#     if now_block.type == "i" and fa_block.type == "i":
#         await handle(now.parent, siyuan)
#     else:
#         await add_note(now.id, siyuan)


tag_attr_name = "ankilink"
assets_replacement = "assets"


def update_siyuan_parser():
    global tag_attr_name
    global assets_replacement
    if "siyuan" not in conf:
        conf["siyuan"] = {}
    tag_attr_name = conf["siyuan"].get("custom_attr_name", "custom-ankilink")
    assets_replacement = conf["siyuan"].get("assets_replacement", "assets")


discovered_notetypes += [SQA, SMQA, SCloze, SListCloze, STableCloze, SWatch]
config_updater.append((update_siyuan_parser, 5))
update_config()


def spec_format(text: str) -> str:
    global assets_replacement
    lines = [x for x in text.splitlines() if x != ""]
    text = ""
    for x in lines:
        if x.startswith(" "):
            # duplicate leading space for lists
            p = 0
            while x[p] == " ":
                p = p + 1
            x = x[:p] + x[:p] + x[p:]
        text = text + x + "\n"
    text = remove_suffix(text, "\n")
    text = text.replace(r"(assets/", f"({assets_replacement}/")
    return text


async def add_note(block: api.SiyuanBlock, note_list: list[api.SiyuanBlock]) -> None:
    # should pass in full block!
    text = spec_format(block.markdown)
    note = markdown.get_note(text, extra_params={"SiyuanID": block.id})
    if note is None:
        return
    note_list.append(note)


async def sync(last_sync_time: str) -> None:
    global tag_attr_name
    siyuan = api.Siyuan(token=conf["siyuan"].get("api_token", ""))
    note_list: list[api.SiyuanBlock] = []
    subs = await siyuan.get_blocks_by_sql(
        f"WHERE parent_id in (SELECT block_id FROM attributes WHERE name='{tag_attr_name}') AND updated>'{last_sync_time}'",
        full=True,
    )
    parents: dict[str, api.SiyuanBlock] = {}
    for x in subs:
        try:
            if x.parent_id not in parents:
                await (await x.parent).ensure()
                parents[x.parent_id] = await x.parent
            attr_conf = await parents[x.parent_id].attrs.get(tag_attr_name, "")
            old_conf = config.parse_config(attr_conf)
            await add_note(x, note_list)
            config.execute_config(old_conf)
        except Exception as e:
            logger.warning(
                f"An error occurred while parsing config.\nSiyuanID:{x.parent_id}\nProperty:\n{attr_conf}\nExpcetion:{e.with_traceback()}"
            )

    await siyuan.close()
    return note_list



from __future__ import annotations

import argparse
import contextlib
import os.path

import toml

from AnkiIn import config as ankiin_config
from AnkiIn.config import dict as conf
from ankisiyuan import config


version_name = "0.1.0"

help_info = {"config": "set your config file path."}

parser = argparse.ArgumentParser(
    description="AnkiSiyuan. Link your data in Siyuan with Anki!"
)
parser.add_argument(
    "-v", "--version", action="version", version=f"Anki Siyuan v{version_name}"
)
parser.add_argument(
    "-c",
    "--config",
    metavar="config",
    default=config.config_path,
    help=help_info["config"],
)


def enable_log_file():
    conf["log_config"]["handlers"]["log_file"] = {
        "level": "DEBUG" if config.log_debug else "INFO",
        "formatter": "standard",
        "class": "logging.FileHandler",
        "filename": "log.txt",
        "mode": "a",
    }
    for x in conf["log_config"]["loggers"].keys():
        conf["log_config"]["loggers"][x]["handlers"].append("log_file")


def parse():
    args = parser.parse_args()
    config.config_path = args.config
    conf["siyuan"] = {}
    conf["siyuan"]["api_token"] = ""
    conf["siyuan"]["custom_attr_name"] = "custom-ankilink"
    conf["siyuan"]["assets_replacement"] = "assets"
    with contextlib.suppress(Exception), open("last_sync_time", encoding="UTF-8") as f:
        config.last_sync_time = f.read()
    enable_log_file()
    load_config_file()
    ankiin_config.update_config()


def load_config_file():
    if os.path.isfile(config.config_path):
        ankiin_config.execute_config(toml.load(config.config_path))
    else:
        print(f"Warning: Config File {config.config_path} doesn't exist!")



from AnkiIn.config import config_updater, dict as conf
from AnkiIn.notetypes.ListCloze import check as super_check

from .SCloze import get as cget


notetype_name = "SListCloze"
if notetype_name not in conf["notetype"]:
    conf["notetype"][notetype_name] = {}
settings = conf["notetype"][notetype_name]

priority = None


def update_siyuan_list_cloze_config():
    global settings, priority

    priority = settings.get("priority", 16)


config_updater.append((update_siyuan_list_cloze_config, 16))


def check(lines: list, extra_params={}) -> bool:
    return super_check(lines=lines, extra_params=extra_params)


def get(text: str, deck: str, tags: list, extra_params={}):
    return cget(text=text, deck=deck, tags=tags, extra_params=extra_params)
